from manhattan.formatters.text import remove_markup

__all__ = ['flows_to_text']


def flows_to_text(flows):
    """Convert HTML within a list of flows to text"""

    # Build a list of text snippets from the HTML contents
    texts = []
    for snippets in flows.values():
        for snippet in snippets:
            for content in snippet['local_contents'].values():
                if type(content) is list:
                    continue

                text = remove_markup(content).strip()
                if text:
                    texts.append(text)

    return ' '.join(texts)



"""
Generic get global snippets chain.
"""

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.content.snippets import GlobalSnippet, SnippetType
from manhattan.manage.views import utils as manage_utils
from manhattan.manage.views import factories as manage_factories
from mongoframes import ASC, In, Q

__all__ = ['global_snippets_chains']


# Define the chains
global_snippets_chains = ChainMgr()

# GET
global_snippets_chains['get'] = Chain([
    'config',
    'authenticate',
    'get_global_snippets',
    'render_json'
    ])


# Define the links
global_snippets_chains.set_link(manage_factories.config())
global_snippets_chains.set_link(manage_factories.authenticate())

@global_snippets_chains.link
def get_global_snippets(state):
    """
    Get a list of global snippets available to the flow.

    This link adds a `global_snippets` list to the state.
    """

    # Get the content flow Id requested
    flow_id = flask.request.args.get('flow', None)
    if not flow_id:
        return manage_utils.json_fail('No `flow` specified')

    # Get a list of types supported by the flow
    snippet_types = SnippetType.get_snippet_types([flow_id])
    type_ids = [t.id for t in snippet_types.values()]

    # Get a list of global snippets available to the flow
    state.global_snippets = GlobalSnippet.many(
        In(Q.type, type_ids),
        sort=[('label', ASC)]
    )

@global_snippets_chains.link
def render_json(state):
    """
    Render and return the global snippet in JSON format.
    """

    # Build the payload
    payload = {
        'snippets': [
            s.to_snippet().to_json_type() for s in state.global_snippets
        ]
    }

    return manage_utils.json_success(payload)


import flask
import jinja2
from manhattan import forms, manage

from manhattan.content.snippets import GlobalRegion, Region
from manhattan.content.views.update_global_contents import \
        update_global_contents_chains
from manhattan.content.utils.templates import (
    image_proxy,
    image_set_proxy,
    image_set_version,
    image_transform
)

__all__ = ['Content']


class ContentPublic:
    """
    The `ContentPublic` class provides initialization code for the package on
    public  applications.
    """

    def __init__(self,
        app,
        authenticate=None,
        base_url='',
        manage_app_tag='manage'
        ):
        self._app = app
        self._manage_app_tag = manage_app_tag

        # Register self with the app
        self._app._content = self

        # Jinja

        # Globals
        app.jinja_env.globals.update(dict(
            get_global_region=GlobalRegion.get_content,
            get_local_region=Region.get_content,
            image_proxy=image_proxy,
            image_set_proxy=image_set_proxy,
            image_set_version=image_set_version,
            image_transform=image_transform
        ))

        # Context processors
        @app.context_processor
        def add_context():

            # Add a context processor to inject information required by the
            # frontend content editing environment into the Jinja environment.
            context = {
                'manhattan_content_init': False,
                'manhattan_content_base_url': None,
                'manhattan_content_var_name': None
            }

            # Check there's a blueprint associated with the URL
            if not flask.request.blueprint:
                return context

            # Find the manage app (there must be a manage app to support
            # content editing).
            manage_app = manage.utils.get_app(
                app._dispatcher,
                manage_app_tag
            )
            if not manage_app:
                return context

            # Ensure a user is logged in
            if authenticate and not authenticate():
                return context

            # The content editing environment should be initialized
            context['manhattan_content_init'] = True

            # Add CSRF token
            if manage_app.config.get('CSRF_SECRET_KEY'):
                context['manhattan_content_csrf_token'] = \
                    forms.CSRF.generate_token()

            # Find the manage blueprint associated with the request blueprint
            blueprint_name = 'manage_' + flask.request.blueprint
            manage_blueprint = manage_app.blueprints.get(blueprint_name)
            if manage_blueprint:

                # Look up the manage blueprint's config and see if it provides
                # a base URL.
                config = manage.utils.get_config(manage_app, blueprint_name)
                if hasattr(config, 'content_base_url'):
                    context['manhattan_content_base_url'] = \
                        config.content_base_url

                # Generate base params required to find the document
                context['manhattan_content_var_name'] = config.var_name

            return context


class ContentManage:
    """
    The `ContentManage` class provides initialization code for the package on
    manage applications.
    """

    def __init__(
        self,
        app,
        base_url='',
        update_global_contents_path='/update-global-contents'
        ):
        self._app = app

        # Views

        # Add view for updating global regions
        self._app.add_url_rule(
            base_url + update_global_contents_path,
            endpoint='update_global_contents',
            view_func=update_global_contents_chains.copy().flask_view(),
            methods=['POST']
            )




import json

import cssutils
import flask
from manhattan.assets import Asset
from manhattan.assets.transforms.base import BaseTransform
from manhattan.assets.transforms.images import Fit, Output

__all__ = [
    'sync_image_fixtures',
    'sync_images',
    'sync_picture_fixtures'
]


def get_default_output_transform():
    """Get the default output transform for images"""

    config = flask.current_app.config

    return Output(
        format=config.get('CONTENT_IMAGE_FORMAT', 'jpg'),
        quality=config.get('CONTENT_IMAGE_QUALITY', 85),
        lossless=config.get('CONTENT_IMAGE_LOSSLESS', False)
    )

def sync_image_fixtures(snippet, soup):
    """Synchronize image fixtures in the HTML soup with the given snippet"""

    asset_mgr = flask.current_app.asset_mgr
    config = flask.current_app.config

    # The name of the variation generated to match the image in the page
    variation_name = config.get('CONTENT_VARIATION_NAME', 'image')

    # A list of image tags
    img_fixtures = soup.find_all(
        **{
            'data-ce-tag': 'img-fixture',
            'data-mh-asset-key': True
        }
    )

    # A table of assets for the snippet
    assets = snippet.assets

    # Build a list of assets to make permanent (persist) and to generate new
    # variations for (transform).
    assets_to_persist = []
    assets_to_transform = []

    for img_fixture in img_fixtures:

        # Find the image within the image fixture
        img = img_fixture.find('img')

        # Get the asset key
        asset_key = img_fixture.get('data-mh-asset-key', '').strip()

        if not asset_key:
            # No valid asset key
            continue

        # Get base and transforms
        base_transforms = json.loads(
            img_fixture.get('data-mh-base-transforms') or '[]'
        )

        # Look for local transforms
        local_transforms = []
        if img_fixture.get('data-mh-local-transforms'):
            local_transforms = json.loads(
                img_fixture['data-mh-local-transforms']
            )
            del img_fixture['data-mh-local-transforms']

        # Determine if we need to persist and/or transform the image
        asset = asset_mgr.get_temporary_by_key(asset_key)
        transform_required = False

        if asset:

            # New image
            assets_to_persist.append(asset)
            transform_required = True
            assets[asset_key] = asset

        elif asset_key in assets:

            # Existing image
            asset = Asset(assets[asset_key])

            if not base_transforms:

                # Legacy support - previously base transforms were cleared
                # from tags, we now keep them to provide support for setting
                # the initial crop/rotate values in the client side image
                # editor to match those currently applied to the image.
                base_transforms = asset.base_transforms

            if json.dumps(base_transforms) != json.dumps(asset.base_transforms):
                transform_required = True

        else:

            # Can't find a matching asset
            continue

        if transform_required:

            if not local_transforms \
                    or local_transforms[-1]['id'] != 'image.output':

                local_transforms \
                        = [get_default_output_transform().to_json_type()]

            assets_to_transform.append((
                asset,
                {
                    variation_name: [
                        BaseTransform.from_json_type(t)
                        for t in local_transforms
                    ]
                },
                [BaseTransform.from_json_type(t) for t in base_transforms]
            ))

    # Persist all new images and apply transforms to all new and modified
    # images.
    asset_mgr.persist_many(assets_to_persist)
    asset_mgr.generate_variations_for_many(
        [a[0] for a in assets_to_transform],
        {a[0].key: a[1] for a in assets_to_transform},
        {a[0].key: a[2] for a in assets_to_transform}
    )

    for img_fixture in img_fixtures:

        # Find the image within the image fixture
        img = img_fixture.find('img')

        # Get the asset key
        asset_key = img_fixture.get('data-mh-asset-key', '').strip()
        asset = assets.get(asset_key)

        if not asset:
            # Asset not in in the assets map for the snippet, ignore it.
            continue

        # Update the image src
        if not isinstance(asset, Asset):
            asset = Asset(asset)

        variation = asset.variations[variation_name]
        if not isinstance(variation, Asset):
            variation = Asset(variation)

        img['src'] = variation.url

        # Update the background image for the fixture
        style = cssutils.parseStyle(img_fixture['style'])
        style['background-image'] = f'url(\'{img["src"]}\')'
        img_fixture['style'] = style.cssText

        # Update the draft URL
        draft = asset.variations['--draft--']
        if not isinstance(draft, Asset):
            draft = Asset(draft)

        img_fixture['data-mh-draft'] = draft.url

    # Update the assets for the snippet
    snippet.assets = assets

def sync_images(snippet, soup):
    """Synchronize the <img> tags in the HTML soup with the given snippet"""

    asset_mgr = flask.current_app.asset_mgr
    config = flask.current_app.config

    # The name of the variation generated to match the image in the page
    variation_name = config.get('CONTENT_VARIATION_NAME', 'image')

    # A list of image tags
    imgs = soup.find_all('img', **{'data-mh-asset-key': True})

    # A table of assets for the snippet
    assets = snippet.assets

    # Build a list of assets to make permanent (persist) and to generate new
    # variations for (transform).
    assets_to_persist = []
    assets_to_transform = []

    for img in imgs:

        # Get the asset key
        asset_key = img.get('data-mh-asset-key', '').strip()

        if not asset_key:
            # No valid asset key
            continue

        # Get base and transforms
        base_transforms = json.loads(
            img.get('data-mh-base-transforms') or '[]'
        )

        # Get the width
        width = int(img['width'])

        # Determine if we need to persist and/or transform the image
        asset = asset_mgr.get_temporary_by_key(asset_key)
        transform_required = False

        if asset:

            # New image
            assets_to_persist.append(asset)
            transform_required = True
            assets[asset_key] = asset

        elif asset_key in assets:

            # Existing image
            asset = Asset(assets[asset_key])

            if not base_transforms:

                # Legacy support - previously base transforms were cleared
                # from tags, we now keep them to provide support for setting
                # the initial crop/rotate values in the client side image
                # editor to match those currently applied to the image.
                base_transforms = asset.base_transforms

            if json.dumps(base_transforms) != json.dumps(asset.base_transforms):
                transform_required = True

            variation = asset.variations.get(variation_name)
            if width != variation['core_meta']['image']['size'][0]:
                transform_required = True

        else:

            # Can't find a matching asset
            continue

        if transform_required:

            local_transforms = [
                Fit(width, 99999),
                get_default_output_transform()
            ]

            assets_to_transform.append((
                asset,
                {variation_name: local_transforms},
                [BaseTransform.from_json_type(t) for t in base_transforms]
            ))

    # Persist all new images and apply transforms to all new and modified
    # images.
    asset_mgr.persist_many(assets_to_persist)
    asset_mgr.generate_variations_for_many(
        [a[0] for a in assets_to_transform],
        {a[0].key: a[1] for a in assets_to_transform},
        {a[0].key: a[2] for a in assets_to_transform}
    )

    for img in imgs:

        # Get the asset key
        asset_key = img.get('data-mh-asset-key', '').strip()
        asset = assets.get(asset_key)

        if not asset:
            # Asset not in in the assets map for the snippet, ignore it.
            continue

        # Update the image src
        if not isinstance(asset, Asset):
            asset = Asset(asset)

        variation = asset.variations[variation_name]
        if not isinstance(variation, Asset):
            variation = Asset(variation)

        img['src'] = variation.url

        # Update the draft URL
        draft = asset.variations['--draft--']
        if not isinstance(draft, Asset):
            draft = Asset(draft)

        img['data-mh-draft'] = draft.url

    # Update the assets for the snippet
    snippet.assets = assets

def sync_picture_fixtures(snippet, soup):
    """Synchronize picture fixtures in the HTML soup with the given snippet"""

    asset_mgr = flask.current_app.asset_mgr
    config = flask.current_app.config

    # A list of image tags
    picture_fixtures = soup.find_all('picture', **{'data-fixture': True})

    # A table of assets for the snippet
    assets = snippet.assets

    # Build a list of assets to make permanent (persist) and to generate new
    # variations for (transform).
    assets_to_persist = []
    assets_to_transform = []

    for picture_fixture in picture_fixtures:

        # Get a list of sources for the picture
        sources = picture_fixture.find_all('source')

        if len(sources) == 0:

            # No image sources
            continue

        if not sources[0].get('data-mh-asset-key', '').strip():

            # Base version has no asset key
            continue

        # Determine the base source version
        base_source = sources[0]

        new_assets = {}
        for source in sources:

            # Get the version for the source
            version = source['data-mh-version']

            # Get the asset key
            asset_key = source.get('data-mh-asset-key', '').strip()

            if not asset_key:
                asset_key = base_source['data-mh-asset-key'].strip()

            asset_version_key = f'{version}:{asset_key}'

            # Get base and transforms
            base_transforms = json.loads(
                source.get('data-mh-base-transforms') or '[]'
            )

            # Look for local transforms
            local_transforms = json.loads(
                source.get('data-mh-local-transforms') or '[]'
            )
            del source['data-mh-local-transforms']

            asset = new_assets.get(
                asset_key,
                asset_mgr.get_temporary_by_key(asset_key)
            )
            transform_required = False

            if asset:

                # New image
                if asset not in assets_to_persist:
                    assets_to_persist.append(asset)
                    new_assets[asset_key] = asset

                transform_required = True
                assets[asset_version_key] = [asset, base_transforms]

            elif asset_version_key in assets:

                # Existing image
                asset, existing_base_transforms = assets[asset_version_key]
                asset = Asset(asset)

                if json.dumps(base_transforms) \
                        != json.dumps(existing_base_transforms):

                    transform_required = True

                assets[asset_version_key] = [asset, base_transforms]

            else:

                # Can't find a matching asset
                continue

            if transform_required:
                assets_to_transform.append((
                    asset,
                    {
                        version:
                            [
                                BaseTransform.from_json_type(t)
                                for t in base_transforms
                            ] + [
                                BaseTransform.from_json_type(t)
                                for t in local_transforms
                            ]
                    },
                    []
                ))

    # Persist all new images and apply transforms to all new and modified
    # images.
    merged_assets_to_transform = {}
    for asset_to_transform in assets_to_transform:
        key = asset_to_transform[0].key
        variations = asset_to_transform[1]

        if key not in merged_assets_to_transform:
            merged_assets_to_transform[key] = {}

        merged_assets_to_transform[key].update(variations)

    asset_mgr.persist_many(assets_to_persist)
    asset_mgr.generate_variations_for_many(
        [a[0] for a in assets_to_transform],
        merged_assets_to_transform,
        {a[0].key: a[2] for a in assets_to_transform}
    )

    for picture_fixture in picture_fixtures:

        # Get a list of sources for the picture
        sources = picture_fixture.find_all('source')

        if len(sources) == 0:

            # No image sources
            continue

        if not sources[0].get('data-mh-asset-key', '').strip():

            # Base version has no asset key
            continue

        # Determine the base source version
        base_source = sources[0]

        for source in sources:

            # Get the version for the source
            version = source['data-mh-version']

            # Get the asset key
            has_own_image = True
            asset_key = source.get('data-mh-asset-key', '').strip()

            if not asset_key:
                has_own_image = False
                asset_key = base_source['data-mh-asset-key'].strip()

            asset_version_key = f'{version}:{asset_key}'

            image_set = assets.get(asset_version_key)

            if not image_set:
                # Asset not in in the assets map for the snippet, ignore it.
                continue

            asset, _ = image_set

            # Update the image src
            if not isinstance(asset, Asset):
                asset = Asset(asset)

            # Set the source URL
            version_variation = asset.variations[version]
            if not isinstance(version_variation, Asset):
                version_variation = Asset(version_variation)

            source['srcset'] = version_variation.url

            if has_own_image:

                # Update the draft URL
                draft = asset.variations['--draft--']
                if not isinstance(draft, Asset):
                    draft = Asset(draft)

                source['data-mh-draft'] = draft.url

            if version == base_source['data-mh-version']:

                img = picture_fixture.find('img')
                img['src'] = source['srcset']

    # Update the assets for the snippet
    snippet.assets = assets



from collections import namedtuple

from .add_snippet import add_snippet_chains
from .add_global_snippet import add_global_snippet_chains
from .change_snippet_scope import change_snippet_scope_chains
from .delete_snippet import delete_snippet_chains
from .global_snippets import global_snippets_chains
from .order_snippets import order_snippets_chains
from .snippet_types import snippet_types_chains
from .snippets import snippets_chains
from .update_contents import update_contents_chains
from .update_flow_contents import update_flow_contents_chains
from .update_global_contents import update_global_contents_chains
from .update_snippet_settings import update_snippet_settings_chains

__all__ = ['generic']


# We name space generic views using a named tuple to provide a slightly nicer
# way to access them, e.g:
#
#     from manhattan.content.views import generic
#
#     view = generic.snippets
#
# And to make it easy to iterate through the list of generic views to make
# changes, e.g:
#
#     def authenticate(state):
#         """A custom authenticator for my site"""
#
#         ...
#
#     for view in generic:
#         view.set_link(authenticate)

# Define the named tuple (preventing the list of generic views being altered)
Generic = namedtuple(
    'Generic',
    [
        'add_snippet',
        'add_global_snippet',
        'change_snippet_scope',
        'delete_snippet',
        'global_snippets',
        'order_snippets',
        'snippet_types',
        'snippets',
        'update_contents',
        'update_flow_contents',
        'update_global_contents',
        'update_snippet_settings'
    ])

# Create an instance of Generic containing all the generic views
generic = Generic(
    add_snippet=add_snippet_chains,
    add_global_snippet=add_global_snippet_chains,
    change_snippet_scope=change_snippet_scope_chains,
    delete_snippet=delete_snippet_chains,
    global_snippets=global_snippets_chains,
    order_snippets=order_snippets_chains,
    snippet_types=snippet_types_chains,
    snippets=snippets_chains,
    update_contents=update_contents_chains,
    update_flow_contents=update_flow_contents_chains,
    update_global_contents=update_global_contents_chains,
    update_snippet_settings=update_snippet_settings_chains
    )


"""
Generic get snippet types chain.
"""

from operator import attrgetter

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.content.snippets import SnippetType
from manhattan.manage.views import utils as manage_utils
from manhattan.manage.views import factories as manage_factories

__all__ = ['snippet_types_chains']


# Define the chains
snippet_types_chains = ChainMgr()

# GET
snippet_types_chains['get'] = Chain([
    'config',
    'authenticate',
    'get_snippet_types',
    'render_json'
    ])


# Define the links
snippet_types_chains.set_link(manage_factories.config())
snippet_types_chains.set_link(manage_factories.authenticate())

@snippet_types_chains.link
def get_snippet_types(state):
    """
    Get a list of snippet types available to the given content flow.

    This link adds a `snippet_types` list to the state.
    """

    # Get the content flow Id requested
    flow_id = flask.request.args.get('flow', None)
    if not flow_id:
        return manage_utils.json_fail('No `flow` specified')

    # Get the available snippet types for the flow
    snippet_types = list(SnippetType.get_snippet_types([flow_id]).values())

    # Order the snippet types by their label
    snippet_types.sort(key=attrgetter('label'))

    state.snippet_types = snippet_types

@snippet_types_chains.link
def render_json(state):
    """Render and return the snippet types in JSON format"""

    # Build the payload
    payload = {'snippet_types': [s.to_json_type() for s in state.snippet_types]}

    return manage_utils.json_success(payload)



"""
Generic update contents chain for content flows.

: `content_flows_field`
    The field within the document that contains the content flows (defaults to
    'flows').

"""

import logging
from copy import deepcopy
import json
import re

from bs4 import BeautifulSoup
import cssutils
import flask
from manhattan.assets import Asset
from manhattan.assets.transforms.base import BaseTransform
from manhattan.assets.transforms.images import Fit, Output
from manhattan.chains import Chain, ChainMgr
from manhattan.content.snippets import Snippet
from manhattan.content.views import factories
from manhattan.manage.views import utils as manage_utils
from manhattan.manage.views import factories as manage_factories
from .utils import sync_image_fixtures, sync_images, sync_picture_fixtures

cssutils.log.setLevel(logging.CRITICAL)


__all__ = ['update_flow_contents_chains']


# Define the chains
update_flow_contents_chains = ChainMgr()

update_flow_contents_chains['post'] = Chain([
    'config',
    'authenticate',
    'get_document',
    'get_flows',
    'get_contents',
    'sync_imagery',
    'update_contents',
    'update_flows',
    'render_json'
])


# Define the links

update_flow_contents_chains.set_link(
    manage_factories.config(content_flows_field='flows')
)
update_flow_contents_chains.set_link(manage_factories.authenticate())
update_flow_contents_chains.set_link(manage_factories.get_document())
update_flow_contents_chains.set_link(factories.get_contents())

@update_flow_contents_chains.link
def get_flows(state):
    """
    Get the flows that we're applying updates to.

    This link adds `flows` and `original_flows` keys to the state containing the
    flows that are to be updated.
    """

    # Get the document we're add the snippet to
    document = state[state.manage_config.var_name]

    # Get the flows we'll be updating
    state.flows = deepcopy(document.get(state.content_flows_field) or {})
    state.original_flows = deepcopy(state.flows)

    # Ensure all snippets in all flows are `Snippet` instances
    for flow_id, flow in state.flows.items():
        for i, snippet in enumerate(flow):
            if not isinstance(snippet, Snippet):
                flow[i] = Snippet(snippet)

@update_flow_contents_chains.link
def sync_imagery(state):
    """
    Temporary and permenant images inserted into the contents must be
    synchronized, by which we mean that temporary assets are converted to
    permenant assets and the content variation of the image is resized to
    match the width/height.
    """

    asset_mgr = flask.current_app.asset_mgr
    config = flask.current_app.config
    variation_name = config.get('CONTENT_VARIATION_NAME', 'image')

    # Build a lookup table of all snippets with the page flows
    snippet_lookup = {}
    for flow_id, flow in state.flows.items():
        for snippet in flow:
            snippet_lookup[snippet.id] = snippet

    # Update the images
    for snippet_id, contents in state.contents.items():
        for region, content in contents.items():

            # Parse the content of the region
            soup = BeautifulSoup(content, 'lxml')

            # Find any element within the content that relates to an asset
            sync_images(snippet_lookup[snippet_id], soup)
            sync_image_fixtures(snippet_lookup[snippet_id], soup)
            sync_picture_fixtures(snippet_lookup[snippet_id], soup)

            # Save the content of the region back to the snippet
            soup.html.unwrap()
            soup.body.unwrap()
            contents[region] = str(soup)

@update_flow_contents_chains.link
def update_contents(state):
    """Update the contents of snippets with the flows"""

    # Apply the changes
    for flow_id, flow in state.flows.items():
        for snippet in flow:
            if snippet.id not in state.contents:
                continue

            if snippet.scope == 'local':
                # Apply changes to this snippet
                snippet.local_contents.update(state.contents[snippet.id])

            elif snippet.scope == 'global':
                # Apply changes to the global snippet this snippet references
                global_snippet = snippet.global_snippet

                # Update the contents
                global_contents = global_snippet.contents.copy()
                global_contents.update(state.contents[snippet.id])

                # Update the global snippet
                global_snippet.logged_update(
                    state.manage_user,
                    {'contents': global_contents}
                )

                # Clear global snippet cache from the snippet
                snippet._global_snippet_cache = None

@update_flow_contents_chains.link
def update_flows(state):
    """Perform a logged or straight update of flows within the document"""

    # Get the document we're add the snippet to
    document = state[state.manage_config.var_name]

    # Check to see if the frame class supports `logged_update`s
    if hasattr(state.manage_config.frame_cls, 'logged_update'):

        # Logged update
        document.logged_update(
            state.manage_user,
            {state.content_flows_field: state.flows}
        )

    else:
        # Straight update (no logging)
        setattr(document, state.content_flows_field, state.flows)
        document.update(state.content_flows_field)

@update_flow_contents_chains.link
def render_json(state):
    """Render a JSON response for the successful saving of content"""
    return manage_utils.json_success()


# Utils

def sync_asset_tags(snippet, tags):
    """
    Inspect a list of asset (image based) tags for changes and update them as
    required.
    """

    asset_mgr = flask.current_app.asset_mgr
    config = flask.current_app.config
    variation_name = config.get('CONTENT_VARIATION_NAME', 'image')

    # Build a look up table of all assets within the snippet
    contents = snippet.local_contents
    if snippet.scope == 'global':
        contents = snippet.global_snippet.contents
    asset_lookup = {a[0]: a[1] for a in contents.get('__assets__', [])}

    for tag in tags:
        base_transforms = []
        local_transforms = []

        # Cater for the scenario where the element is an image fixture
        img = tag
        if tag.name != 'img':
            img = tag.find('img')

        # Make sure the key has a value
        key = tag['data-mh-asset-key'].strip()
        if not key:
            continue

        # Look for base transforms
        if tag.get('data-mh-base-transforms'):
            base_transforms = json.loads(tag['data-mh-base-transforms'])
            del tag['data-mh-base-transforms']

        # Check for the width of the image, if this has changed it will form
        # part of the local transforms.
        new_width = None
        if img.get('width', '').isdigit() and int(img['width']) > 0:
            new_width = int(img['width'])

        # Look for local transforms
        if tag.get('data-mh-local-transforms'):
            local_transforms = json.loads(tag['data-mh-local-transforms'])
            del tag['data-mh-local-transforms']

        # Find the asset
        if asset_mgr.get_temporary_by_key(key):

            # New asset (store it)
            asset = asset_mgr.get_temporary_by_key(key)
            flask.current_app.asset_mgr.store(asset)

        elif key in asset_lookup:

            # Existing asset
            asset = Asset(asset_lookup[key])

            if base_transforms:

                # If the base transforms are the same as the existing
                # transforms then set the transforms to an empty list.
                new = json.dumps(base_transforms)
                old = json.dumps(asset.base_transforms)
                if new == old:
                    base_transforms = []

            if new_width and not base_transforms:

                # If the new width is the same as the existing width then set
                # width to None.
                variation = asset.variations.get(variation_name)
                if new_width == variation['core_meta']['image']['size'][0]:
                    new_width = None

        else:
            # Asset not found for the given key (ignore the tag)
            continue

        # If the image has a new width then create a local transform for it
        if new_width:
            local_transforms.append(Fit(new_width, 99999).to_json_type())

        # If there are base or local transforms required for the image then
        # generate the require variation.
        if base_transforms or local_transforms:

            # If there are no base transforms check if there are transforms
            # against the asset.
            if not base_transforms:
                base_transforms = asset.base_transforms

            # Add the standard output quality for images to the local
            # transforms if not output quality is set.
            has_output = False
            for local_transform in local_transforms:
                if local_transform['id'] == 'image.output':
                    has_output = True

            if not has_output:
                local_transforms.append(
                    Output(
                        config.get('CONTENT_IMAGE_FORMAT', 'jpg'),
                        config.get('CONTENT_IMAGE_QUALITY', 85)
                    ).to_json_type()
                )

            asset_mgr.generate_variations(
                asset,
                {
                    variation_name: [BaseTransform.from_json_type(t)
                            for t in local_transforms]
                },
                [BaseTransform.from_json_type(t) for t in base_transforms]
            )

        # Update the image URL
        if not isinstance(asset.variations[variation_name], Asset):
            asset.variations[variation_name] = Asset(
                asset.variations[variation_name]
            )

        img['src'] = asset.variations[variation_name].url
        if tag != img and tag.name != 'picture':

            # Against image fixtures the URL also needs to be set against the
            # tags background image style.
            style = cssutils.parseStyle(tag['style'])
            del style['background-image']
            style['background-image'] = "url('{0}')".format(img['src'])
            tag['style'] = style.cssText

        # Update the draft URL
        if not isinstance(asset.variations['--draft--'], Asset):
            asset.variations['--draft--'] = Asset(
                asset.variations['--draft--']
            )
        tag['data-mh-draft'] = asset.variations['--draft--'].url

        # Store the updated asset information back in the look up table and
        # snippet.
        asset_lookup[key] = asset.to_json_type()
        contents['__assets__'] = [[k, v] for k, v in asset_lookup.items()]



"""
Generic update contents for documents with a field containing content regions.

In some cases a fixed layout is preferable to the more flexible flow layout,
or you may wish to allow users to update more than one field within a document
via the in-page editor. This generic view is designed to provide a basis for
such cases (it's also possible to combine both flow and region based update
views).

Image syncing is only supported against the `content_regions_field`, assets
are stored within the regions field under the special `__assets__` key.

: `content_regions_field`
    The field within the document that contains the content flows (defaults to
    'flows').

"""

import json
import re

from bs4 import BeautifulSoup
import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.content.snippets import Region
from manhattan.content.views import factories
from manhattan.manage.views import utils as manage_utils
from manhattan.manage.views import factories as manage_factories
from mongoframes import Q
from .utils import sync_image_fixtures, sync_images, sync_picture_fixtures

__all__ = ['update_contents_chains']


# Define the chains
update_contents_chains = ChainMgr()

update_contents_chains['post'] = Chain([
    'config',
    'authenticate',
    'get_document',
    'get_contents',
    'get_regions',
    'sync_imagery',
    'update_document',
    'render_json'
])


# Define the links

update_contents_chains.set_link(
    manage_factories.config(content_regions_field='regions')
)
update_contents_chains.set_link(manage_factories.authenticate())
update_contents_chains.set_link(manage_factories.get_document())
update_contents_chains.set_link(factories.get_contents())

@update_contents_chains.link
def get_regions(state):
    """
    Get copies of the regions within the document we'll be updating.

    This link adds the `regions` key to the state, which contains a dictionary
    of existing regions contents.
    """

    # Get the document we're add the snippet to
    document = state[state.manage_config.var_name]

    # Ensure there's a table to get/set contents against
    assert state.content_regions_field, 'No `content_regions_field` configured'
    regions = document.get(state.content_regions_field) or {}
    setattr(document, state.content_regions_field, regions)
    state.regions = regions.copy()

    # Ensure there's a base region and assets table for each of the content
    # updates.
    for region_name, content in state.contents.items():
        assets = {}
        if region_name in state.regions:
            assets = state.regions.get('assets', {})
        state.regions[region_name] = Region(
            assets=assets,
            content=content
        )

@update_contents_chains.link
def sync_imagery(state):
    """
    Temporary and permenant images inserted into the contents must be
    synchronized, by which we mean that temporary assets are converted to
    permenant assets and the content variation of the image is resized to
    match the width/height.
    """

    # Get the document we're add the snippet to
    document = state[state.manage_config.var_name]

    asset_mgr = flask.current_app.asset_mgr
    config = flask.current_app.config
    variation_name = config.get('CONTENT_VARIATION_NAME', 'image')

    # Update the images
    for region, content in state.contents.items():

        # Parse the content of the region
        soup = BeautifulSoup(content, 'lxml')

        # Find any element within the content that relates to an asset
        sync_images(state.regions[region], soup)
        sync_image_fixtures(state.regions[region], soup)
        sync_picture_fixtures(state.regions[region], soup)

        # Save the content of the region back to the snippet
        soup.html.unwrap()
        soup.body.unwrap()
        state.regions[region].content = str(soup)

@update_contents_chains.link
def update_document(state):
    """Update documents contents"""

    # Get the document we're updating the contents of
    document = state[state.manage_config.var_name]

    # Update the document
    document.logged_update(
        state.manage_user,
        {state.content_regions_field: state.regions}
    )

@update_contents_chains.link
def render_json(state):
    """Render a JSON response for the successful saving of content"""
    return manage_utils.json_success()




from jinja2 import Markup
from jinja2.utils import htmlsafe_json_dumps

__all__ = [
    'image_proxy',
    'image_set_proxy',
    'image_set_version',
    'image_transform'
]


def image_proxy(
    transforms,
    crop_aspect_ratio=None,
    fix_crop_aspect_ratio=False
):
    """
    Shortcut for creating data attributes for an image proxy.

    Usage:

        {{ image_proxy(
            transforms=[
                image_transform(id='fit', width=500),
                image_transform(id='output', format='webp', quality=95)
            ],
            crop_aspect_ratio=1.5,
            fixed_crop=False
        ) }}

    """

    attrs = {
        'data-mh-image-proxy': True,
        'data-mh-image-transforms': htmlsafe_json_dumps(transforms)
    }

    if crop_aspect_ratio:
        attrs['data-mh-image-crop-aspect-ratio'] = crop_aspect_ratio

    if fix_crop_aspect_ratio:
        assert crop_aspect_ratio, (
            'A fixed crop aspect ratio can only be applied if a crop aspect '
            'ratio is given.'
        )

        attrs['data-mh-image-fixed-crop'] = True

    return Markup(
        '\n'.join([
            k if v is True else '='.join([k, f"'{v}'"])
            for k, v in attrs.items()
        ])
    )

def image_set_proxy(*versions, fix_crop_aspect_ratio=False):
    """
    Shortcut for creating data attributes for an image set proxy.

    Usage:

        {{ image_set_proxy(
            image_set_version(
                version='l',
                label='Desktop',
                transforms=[
                    image_transform(id='fit', width=1000),
                    image_transform(id='output', format='webp', quality=95)
                ],
                media="(min-width: 960px)",
                crop_aspect_ratio=1.5
            ),
            image_set_version(
                version='m',
                label='Tablet',
                transforms=[
                    image_transform(id='fit', width=750),
                    image_transform(id='output', format='webp', quality=85)
                ],
                media="(min-width: 640px)",
                crop_aspect_ratio=1.333
            ),
            image_set_version(
                version='s',
                label='Mobile',
                transforms=[
                    image_transform(id='fit', width=400),
                    image_transform(id='output', format='webp', quality=75)
                ],
                media="(min-width: 320px)",
                crop_aspect_ratio=1.0
            )
            fix_crop_aspect_ratio=True
        ) }}

    """

    version_names = [v['version'] for v in versions]
    version_labels = {v['version']: v['label'] for v in versions}
    transforms = {v['version']: v['transforms'] for v in versions}
    media = {v['version']: v['media'] for v in versions}

    attrs = {
        'data-mh-image-set-proxy': True,
        'data-mh-image-set-versions': htmlsafe_json_dumps(version_names),
        'data-mh-image-set-version-labels': htmlsafe_json_dumps(version_labels),
        'data-mh-image-set-transforms': htmlsafe_json_dumps(transforms),
        'data-mh-image-set-media': htmlsafe_json_dumps(media)
    }

    crop_aspect_ratios = None
    if versions[0].get('crop_aspect_ratio'):
        crop_aspect_ratios = {
            v['version']: v['crop_aspect_ratio'] for v in versions
        }
        attrs['data-mh-image-set-crop-aspect-ratios'] = \
                htmlsafe_json_dumps(crop_aspect_ratios)

    if fix_crop_aspect_ratio:
        assert crop_aspect_ratios, (
            'A fixed crop aspect ratio can only be applied if crop aspect '
            'ratios are given.'
        )
        attrs['data-mh-image-set-fix-crop-aspect-ratio'] = True

    return Markup(
        '\n'.join([
            k if v is True else '='.join([k, f"'{v}'"])
            for k, v in attrs.items()
        ])
    )

def image_set_version(
    version,
    label,
    transforms,
    media,
    crop_aspect_ratio=None
):
    """Shortcut for defining an image version within an image set"""

    version = {
        'version': version,
        'label': label,
        'transforms': transforms,
        'media': media
    }

    if crop_aspect_ratio:
        version['crop_aspect_ratio'] = crop_aspect_ratio

    return version

def image_transform(id, **settings):
    """Shortcut for creating image transforms"""

    if '.' not in id:
        id = f'image.{id}'

    return {'id': id, 'settings': settings}




"""
Generic add snippet chain.

: `content_flows_field`
    The field within the document that contains the content flows (defaults to
    'flows').

"""

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.content.snippets import Snippet, SnippetType
from manhattan.content.views import factories
from manhattan.manage.views import utils as manage_utils
from manhattan.manage.views import factories as manage_factories

__all__ = ['add_snippet_chains']


# Define the chains
add_snippet_chains = ChainMgr()

# POST
add_snippet_chains['post'] = Chain([
    'config',
    'authenticate',
    'get_document',
    'get_flow',
    'get_snippet_type',
    'add_snippet',
    'update_flow',
    'render_json'
])


# Define the links
add_snippet_chains.set_link(
    manage_factories.config(content_flows_field='flows')
)
add_snippet_chains.set_link(manage_factories.authenticate())
add_snippet_chains.set_link(manage_factories.get_document())
add_snippet_chains.set_link(factories.get_flow())
add_snippet_chains.set_link(factories.update_flow())

@add_snippet_chains.link
def get_snippet_type(state):
    """
    Get the type of snippet we'll be adding to the flow.

    This link adds `snippet_type` to the state.
    """

    # Get the snippet type Id requested
    snippet_type_id = flask.request.values.get('snippet_type', None)
    if not snippet_type_id:
        return manage_utils.json_fail('No `snippet_type` specified')

    # Get the available snippet types for the flow
    snippet_types = SnippetType.get_snippet_types([state.flow_id])

    # Make sure the given snippet type is available
    snippet_type = snippet_types.get(snippet_type_id, None)
    if not snippet_type:
        return manage_utils.json_fail('Snippet type not found')

    state.snippet_type = snippet_type

@add_snippet_chains.link
def add_snippet(state):
    """
    Add a snippet to the named flow.

    This link adds `snippet` to the state.
    """

    # Get the document we're add the snippet to
    document = state[state.manage_config.var_name]

    # Create the new snippet
    snippet = Snippet(
        type=state.snippet_type.id,
        scope='local',
        local_settings=state.snippet_type.get_default_settings()
    )

    # Add the snippet to the flow
    state.flow.append(snippet)

    # Add the snippet to the state
    state.snippet = snippet

@add_snippet_chains.link
def render_json(state):
    """Render the new snippet as HTML and return it in a JSON response"""

    # Get the document we're add the snippet to
    document = state[state.manage_config.var_name]

    # Render the snippet as HTML
    snippet_html = state.snippet.render(
        {state.manage_config.var_name: document}
    )

    return manage_utils.json_success({'html': snippet_html})



"""
Generic delete snippet chain.

: `content_flows_field`
    The field within the document that contains the content flows (defaults to
    'flows').

"""

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.content.views import factories
from manhattan.manage.views import utils as manage_utils
from manhattan.manage.views import factories as manage_factories

__all__ = ['delete_snippet_chains']


# Define the chains
delete_snippet_chains = ChainMgr()

# POST
delete_snippet_chains['post'] = Chain([
    'config',
    'authenticate',
    'get_document',
    'get_flow',
    'get_snippet',
    'delete_snippet',
    'update_flow',
    'render_json'
])


# Define the links
delete_snippet_chains.set_link(
    manage_factories.config(content_flows_field='flows')
)
delete_snippet_chains.set_link(manage_factories.authenticate())
delete_snippet_chains.set_link(manage_factories.get_document())
delete_snippet_chains.set_link(factories.get_flow())
delete_snippet_chains.set_link(factories.get_snippet())
delete_snippet_chains.set_link(factories.update_flow())

@delete_snippet_chains.link
def delete_snippet(state):
    """Delete a snippet from a flow"""
    state.flow = [s for s in state.flow if s['id'] != state.snippet.id]

@delete_snippet_chains.link
def render_json(state):
    """Render a JSON response for the successful deletion of the snippet"""
    return manage_utils.json_success()



"""
Link factories for content views.
"""

from copy import deepcopy
import json

import flask
from manhattan.content.snippets import Snippet
from manhattan.manage.views import utils as manage_utils
import re

__all__ = [
    'get_contents',
    'get_flow',
    'get_snippet',
    'update_flow'
    ]


def get_contents():
    """
    Return a function that will retreive and deserialize a dictionary of
    content that will be used to update a document (see `update_contents.py`,
    `update_flow_contents.py` and `update_global_contents`).

    This link adds the `contents` key to the state, which contains a dictionary
    of content to apply in the update.
    """

    def get_contents(state):

        # Get the contents we need to update the flows with
        try:
            state.contents = json.loads(
                flask.request.values.get('contents', '{}')
            )
        except ValueError:
            return manage_utils.json_fail('`contents` is not valid JSON')

    return get_contents

def get_flow():
    """
    Return a function to look up a content flow based on the `flow` request
    parameter and add it to the state.

    This link adds `flow` and `flow_id` and `original_flow` to the state. The
    original flow value is used for comparison when making logged updates are
    made.
    """

    def get_flow(state):

        # Get the document were holding the content flow
        document = state[state.manage_config.var_name]

        # Get the content flow Id requested
        flow_id = flask.request.values.get('flow', None)
        if not flow_id:
            return manage_utils.json_fail('No `flow` specified')

        # Ensure there's a table to get/set content flows against
        assert state.content_flows_field, 'No `content_flows_field` configured'
        flows = document.get(state.content_flows_field) or {}
        setattr(document, state.content_flows_field, flows)

        # Ensure all snippets are converted to instances of `Snippet`
        flow = flows.get(flow_id, [])
        for i, snippet in enumerate(flow):
            if not isinstance(snippet, Snippet):
                snippet = Snippet(snippet)
            flow[i] = snippet

        # Add flow and flow Id to the state
        state.original_flow = deepcopy(flow)
        state.flow = deepcopy(flow)
        state.flow_id = flow_id

    return get_flow

def get_snippet():
    """
    Return a function to find a snippet in a flow and add it to the state.

    This link adds `snippet` and to the state, it expects the state to contain
    the `flow` attribute.
    """

    def get_snippet(state):

        # Get the snippet Id to remove
        snippet_id = flask.request.values.get('snippet', None)
        if not snippet_id:
            return manage_utils.json_fail('No `snippet` specified')

        # Attempt to find the snippet
        snippet = None
        for s in state.flow:
            if s['id'] == snippet_id:
                snippet = s
                break

        # Check that the snippet was found
        if not snippet:
            return manage_utils.json_fail('Snippet not found')

        # Add the snippet to the state
        state.snippet = snippet

    return get_snippet

def update_flow():
    """
    Return a function to perform a logged or straight update of the flows field
    within a document for an individual flow.
    """

    def update_flow(state):

        # Get the document we're add the snippet to
        document = state[state.manage_config.var_name]

        # Update the flow within the document
        document[state.content_flows_field][state.flow_id] = state.flow

        # Check to see if the frame class supports `logged_update`s
        if hasattr(state.manage_config.frame_cls, 'logged_update'):

            # Timestamp
            document.__class__.timestamp_update(document, [document])

            # Update the document
            document.update('modified', state.content_flows_field)

            # Create an entry and perform a diff
            entry = document.__class__._change_log_cls({
                'type': 'UPDATED',
                'documents': [document],
                'user': state.manage_user
                })
            entry.add_diff(
                {state.content_flows_field: state.original_flow},
                {state.content_flows_field: state.flow}
            )

            # Log any differences
            if entry.is_diff:
                entry.insert()

        else:
            # Straight update (no logging)
            document.update(state.content_flows_field)

    return update_flow


"""
Generic update global contents asset chain.

This view updates contents with the `GlobalRegion` collection.
"""

import json
import re

from bs4 import BeautifulSoup
import cssutils
import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.content.snippets import GlobalRegion
from manhattan.content.views import factories
from manhattan.manage.views import utils as manage_utils
from manhattan.manage.views import factories as manage_factories
from mongoframes import Q
from .utils import sync_image_fixtures, sync_images, sync_picture_fixtures

__all__ = ['update_global_contents_chains']


# Define the chains
update_global_contents_chains = ChainMgr()

update_global_contents_chains['post'] = Chain([
    'config',
    'authenticate',
    'get_contents',
    'get_global_regions',
    'sync_imagery',
    'update_contents',
    'render_json'
])


# Define the links

update_global_contents_chains.set_link(manage_factories.config())
update_global_contents_chains.set_link(manage_factories.authenticate())
update_global_contents_chains.set_link(factories.get_contents())

@update_global_contents_chains.link
def get_global_regions(state):
    """
    Get the global regions that will be update based on the contents.

    This link adds the `global_regions` key to the state, which contains a
    dictionary of global regions to apply the update to, e.g
    `{region_name: instance}`.

    NOTE: If the contents contains a region name that doesn't exist a new
    region will be created.
    """

    # Build the table of global regions
    global_regions = {}
    for region_name, content in state.contents.items():

        # Attempt to find the named global region
        global_region = GlobalRegion.one(Q.name == region_name)

        # If there's no existing region with that name then add one
        if not global_region:
            global_region = GlobalRegion(name=region_name)
            global_region.logged_insert(state.manage_user)

        global_regions[region_name] = global_region

    # Add the global regions to the state
    state.global_regions = global_regions

@update_global_contents_chains.link
def sync_imagery(state):
    """
    Temporary and permenant images inserted into the contents must be
    synchronized, by which we mean that temporary assets are converted to
    permenant assets and the content variation of the image is resized to
    match the width/height.
    """

    asset_mgr = flask.current_app.asset_mgr
    config = flask.current_app.config
    variation_name = config.get('CONTENT_VARIATION_NAME', 'image')

    # Update the images
    for region, content in state.contents.items():

        # Parse the content of the region
        soup = BeautifulSoup(content, 'lxml')

        # Find any element within the content that relates to an asset
        sync_images(state.global_regions[region], soup)
        sync_image_fixtures(state.global_regions[region], soup)
        sync_picture_fixtures(state.global_regions[region], soup)

        # Save the content of the region back to the snippet
        soup.html.unwrap()
        soup.body.unwrap()
        state.global_regions[region].content = str(soup)

@update_global_contents_chains.link
def update_contents(state):
    """Update the contents of global regions within the contents"""
    for region_name, content in state.contents.items():
        state.global_regions[region_name].logged_update(
            state.manage_user,
            {
                'content': state.global_regions[region_name].content,
                'assets': state.global_regions[region_name].assets
            }
        )

@update_global_contents_chains.link
def render_json(state):
    """Render a JSON response for the successful saving of content"""
    return manage_utils.json_success()



"""
Classes for prototyping and storing content snippets.
"""

import copy
import inspect
import os
import uuid

import flask
import inflection
from manhattan import comparable
from manhattan import manage
from mongoframes import ASC, Frame, IndexModel, Q, SubFrame

__all__ = [
    'GlobalRegion',
    'GlobalSnippet',
    'Region,'
    'Snippet',
    'SnippetType'
]


class GlobalRegion(comparable.ComparableFrame):
    """
    Global regions provide support for named regions of editable content that
    can be included in HTML content.
    """

    _fields = {
        'name',
        'content',
        'assets'
    }

    _indexes = [
        IndexModel([('name', ASC)], unique=True)
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.assets is None:
            self.assets = {}

    @classmethod
    def get_content(cls, name):
        """Return the content for the named global region"""
        global_region = cls.one(Q.name == name)
        if global_region:
            return global_region.content


class GlobalSnippet(comparable.ComparableFrame):
    """
    A snippet that is typically shared across multiple content locations
    allowing updates to be applied globally to all instances of the snippet.
    """

    _fields = {
        'type',
        'visible',

        # Fields only assigned to local snippets
        'settings',
        'contents',

        # Fields only assigned to global snippets
        'label'
    }

    @property
    def snippet_type_cls(self):
        return SnippetType.get_snippet_types().get(self.type)

    def to_snippet(self):
        """Return a `Snippet` instance representing the global snippet"""
        snippet = Snippet(
            id=None,
            type=self.type,
            scope='global',
            global_id=self._id
        )

        # To improve performance we manually set the value of the global snippet
        # cache against the snippet to prevent an unnecessary look up.
        snippet._global_snippet_cache = self

        return snippet


class Region(SubFrame):
    """
    A locally editable region within a page.
    """

    _fields = {
        'content',
        'assets'
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.assets is None:
            self.assets = {}

    @classmethod
    def get_content(cls, regions, name):
        """Return the content for the named global region"""
        return (regions or {}).get(name, {}).get('content')


class Snippet(SubFrame):
    """
    A snippet within a content flow.
    """

    _fields = {
        'id',
        'type',
        'scope',

        # Fields only assigned to local snippets
        'local_settings',
        'local_contents',

        # Fields only assigned to global snippets
        'global_id'
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Defaults
        if self.id is None:
            self.id = str(uuid.uuid4())

        if self.local_settings is None:
            self.local_settings = {}

        if self.local_contents is None:
            self.local_contents = {}

        # A cached version of the associated global snippet (if any)
        self._global_snippet_cache = None

    # Read-only

    @property
    def assets(self):
        return {a[0]: a[1] for a in self.contents.get('__assets__', [])}

    @assets.setter
    def assets(self, value):
        assets = [[k, v] for k, v in value.items()]
        if self.scope == 'global':
            self.global_snippet.contents['__assets__'] = assets
        else:
            self.local_contents['__assets__'] = assets

    @property
    def contents(self):
        if self.scope == 'global':
            return self.global_snippet.contents
        return self.local_contents

    @property
    def global_label(self):
        if self.scope == 'global':
            return self.global_snippet.label

    @property
    def global_snippet(self):
        # Make sure the snippet is global
        if self.scope != 'global':
            return None

        # Ensure we cache the global snippet the first time we request it
        if self._global_snippet_cache is None:
            self._global_snippet_cache = GlobalSnippet.by_id(self.global_id)

        return self._global_snippet_cache

    @property
    def settings(self):
        if self.scope == 'global':
            return self.global_snippet.settings
        return self.local_settings

    @property
    def type_cls(self):
        return SnippetType.get_snippet_types().get(self.type)

    # Methods

    def copy(self):
        """Return a copy of the snippet"""
        return self.__class__(
            id=str(uuid.uuid4()),
            type=self.type,
            scope=self.scope,
            local_settings=copy.deepcopy(self.local_settings),
            local_contents=copy.deepcopy(self.local_contents),
            global_id=self.global_id
        )

    def render(self, template_args=None):
        """Render and return a HTML string for the snippet"""
        return self.type_cls.render(self, template_args)

    def to_json_type(self):
        """Return a JSON safe dictionary representing the snippet"""
        json_data = {
            'id': self.id,
            'type': self.type_cls.to_json_type(),
            'scope': self.scope,
            'settings': self.settings,
            'contents': self.contents
        }

        if self.scope == 'global':
            json_data.update({
                'global_id': str(self.global_id),
                'global_label': self.global_label
            })

        return json_data


class SnippetTypeMeta(type):
    """
    Meta class for snippet types to provide base class attribute defaults and
    automate registration.
    """

    def __new__(meta, name, bases, dct):

        # Defaults
        if 'available' not in dct:
            dct['available'] = True

        if 'flows' not in dct:
            dct['flows'] = set(['*'])

        if 'id' not in dct:
            dct['id'] = inflection.underscore(name)

        if 'label' not in dct:
            label = inflection.underscore(name).replace('_', ' ').capitalize()
            dct['label'] = label

        cls = super().__new__(meta, name, bases, dct)

        if 'template_path' not in dct:
            filepath = os.path.relpath(inspect.getfile(cls))
            parts = os.path.split(filepath)[0].split('/')
            template_name = inflection.underscore(name)
            cls.template_path = os.path.join(
                '/'.join([parts[-1], 'snippet_types']),
                template_name + '.html'
            )

        # Register the class
        if len(bases):
            bases[0]._snippet_types[cls.id] = cls

        return cls


class SnippetType(metaclass=SnippetTypeMeta):
    """
    Snippets are created based on a snippet type, the type determines the
    template used to render the snippet, the settings that can be set against
    the snippet and the behaviour when rendering the template.

    All snippet types must inherit from this class but it should not be used
    directly.

    Snippet types should be static classes, they should not be initialized.
    """

    # A table of all registered snippets
    _snippet_types = {}

    # Flag indicating if the snippet type is currently available, e.g if a
    # snippet type is used only as a base class or if the type has been achived.
    available = False

    # A list of flow rules used to determine which flows the snippet type is
    # available in. Rules support:
    #
    # - exact match ('foo')
    # - endswith ('foo*')
    # - excludes ('^foo')
    # - excludes endswith ('^foo*')
    flows = None

    # A unique Id for the snippet type
    id = ''

    # A verbose label for the snippet type (used to help make the snippet type
    # easy to identify).
    label = None

    # URL to a preview image for the snippet type
    image_url = None

    # Path to the template that will be used when rendering the snippet
    template_path = None

    @classmethod
    def get_default_settings(cls):
        """Return the default settings for a snippet of this type"""
        return {}

    @classmethod
    def get_image_url(cls):
        """Return an image URL for the snippet"""

        if not cls.image_url:
            return ''

        # Get the static asset loader from the Jinja env
        get_static_asset = flask\
            .current_app\
            .jinja_env\
            .globals\
            .get('get_static_asset')

        if not get_static_asset:
            return cls.image_url

        return get_static_asset(cls.image_url)

    @classmethod
    def get_jinja_env(cls):
        """
        Return an environment to render templates with (this method should
        be overridden if you want to provide a custom environment).
        """
        return flask.current_app._dispatcher.app.jinja_env

    @classmethod
    def get_settings_form(cls):
        """Return the form used to set settings a snippet of this type"""
        return None

    @classmethod
    def get_snippet_types(cls, flows=None, only_available=True):
        """
        Return a table of snippet type clases, by default only available snippet
        types are returned (table structure is `{id: cls}). Optionally you can
        specify a list/set of flow Ids to filter by.
        """
        snippet_types = {}
        for id, cls in SnippetType._snippet_types.items():

            if only_available and not cls.available:
                continue

            if flows:

                # Separate the flow rules for the class into lists we can test
                # against.
                exact = [
                    f for f in cls.flows
                    if not (f.endswith('*') or f.startswith('^'))
                ]
                ends = [f[:-1] for f in cls.flows if f.endswith('*')]
                excludes_exact = [f[1:] for f in cls.flows if f.startswith('^')]
                excludes_ends = [
                    f[1:-1] for f in cls.flows
                    if f.endswith('*') and f.startswith('^')
                ]

                # Check if the flow matches one of the class' flow rules
                match = None
                for f in flows:

                    # Excludes exact
                    if match is None:
                        for rule in excludes_exact:
                            if f == rule:
                                match = False
                                break

                    # Excludes ends
                    if match is None:
                        for rule in excludes_ends:
                            if f.startswith(rule):
                                match = False
                                break

                    # Exact
                    if match is None:
                        for rule in exact:
                            if f == rule:
                                match = True
                                break

                    # Ends
                    if match is None:
                        for rule in ends:
                            if f.startswith(rule):
                                match = True
                                break

                if not match:
                    continue

            snippet_types[id] = cls

        return snippet_types

    @classmethod
    def render(cls, snippet, template_args=None):
        """Render and return a HTML string for the given snippet"""
        template = cls.get_jinja_env().get_template(cls.template_path)
        template_args = template_args or {}
        template_args.update({'snippet': snippet})
        return template.render(**template_args)

    @classmethod
    def to_json_type(cls):
        """Return a JSON safe dictionary representing the snippet type"""
        return {
            'id': cls.id,
            'label': cls.label,
            'image_url': cls.get_image_url()
        }
